{$MODE Delphi}

program MP3WAV;

{ XAudio/FPC sample program. Version 1.00
  Shows how to decode a MP3 file and save it's contents in a WAV file

  by Udo Giacomozzi <ugcm@usa.net> or http://come.to/jampy

  May be used freely for non-commercial purposes as long this message
  is not deleted.

}

{$DEFINE ShowSpeed}  // if defined, shows decoding speed

uses
  DOS, XAudio;

{$IFDEF ShowSpeed}
CONST
  TickFreq = 1193180/65536;  // BIOS tick frequency

FUNCTION GetBiosTicks:DWORD; // returns BIOS ticks counter
BEGIN
  GetBiosTicks:=MemL[Seg0040:$6C];
END;
{$ENDIF}


{ Sets the default file extension if none given
    FName  - the file name to be modified
    DefExt - Default extension (format ".XXX" !!) }
function SetDefaultExt(const FName, DefExt:String):String;
var
  fp,fn,fe : String;  // file path/name/extension
begin
  FSplit(FName, fp, fn, fe);
  If fe = '' then fe := DefExt;
  Result := fp+fn+fe;
end;

{ Returns only the name (no path, no extension) of a file name
    FName  - The file name to be modified }
function GetFName(const FName:String):String;
var
  fp,fn,fe : String;  // file path/name/extension
begin
  FSplit(FName, fp, fn, fe);
  Result := fn;
end;

(*** XAudio notification handlers ***)
procedure Notify_Progress(client:Pointer; source:LongInt; code  :LongInt;
                          value :LongInt; message:PChar); cdecl;
begin
  { nothing to do }
end;

procedure Notify_Debug(client:Pointer; source:LongInt; level:LongInt;
                       message:PChar; reason:PChar); cdecl;
begin
  { nothing to do }
end;

procedure Notify_Error(client:Pointer; source:LongInt; code:LongInt;
                       message:PChar; reason:PChar); cdecl;
begin
  { nothing to do }
end;

type { WAV file structures }
  tFileHeader  = packed record
                   ID    : Array[1..4] of Char; {RIFF}
                   Len   : DWORD;
                 end;
  tWavHeader   = Array[1..4] of Char; {WAVE}
  tFormatChunk = packed record
                   ID               : Array[1..4] of Char; {fmt }
                   Len              : DWORD;
                   wFormatTag       : Word;
                   wChannels        : Word;
                   dwSamplesPerSec  : DWord;
                   dwAvgBytesPerSec : DWord;
                   wBlockAlign      : Word;
                   wBitsPerSample   : Word;
                 end;
  tDataChunk   = packed record
                   ID    : Array[1..4] of Char; {data}
                   Len   : DWORD;
                 end;

var
  InFileName   : String;  // name of the input file (.MP3)
  OutFileName  : String;  // name of the output file (.WAV)
  OutF         : file;    // output file
  CharBuf      : Array[Byte] of Char; // char buffer for PChar operations

  WavHead      : record
                   FileHeader : tFileHeader;
                   Wavheader  : tWavHeader;
                   Format     : tFormatChunk;
                   Data       : tDataChunk;
                 end;

  Status       : Longint; // XAudio return code
  decoder      : PXA_DecoderInfo; // XAudio decoder instance
  Module       : XA_InputModule;  // Input/output module handle
  IgnoreErr    : Boolean; // TRUE: Ignore decoding error (> decoding section)
  DecLen       : DWORD;   // MP3 file size in seconds
  DecPos       : Real;    // MP3 file position in seconds
  Channels     : Byte;    // number of channels (1=MONO, 2=STEREO)

{$IFDEF ShowSpeed}
  Timer        : DWORD;   // timer start position
  DecodingTime : DWORD;   // decoding time in milliseconds
{$ENDIF}

begin
  writeln('MP3 to WAV converter using MpegTV''s XAudio decoding engine. Version 1.0');
  writeln('Copyright 1999 Udo Giacomozzi <ugcm@usa.net> [http://come.to/jampy]');
  writeln('Using XAudio decoder version ', decoder_version(nil,nil,nil));
  writeln;

  { Read commandline parameters ------------------------------------------- }
  if ParamCount<1 then
  begin
    writeln('Usage:');
    writeln('  MP3WAV InputFile[.MP3] [OutputFile[.WAV]]');
    Halt;
  end;
  InFileName  := SetDefaultExt(ParamStr(1), '.MP3');

  OutFileName := ParamStr(2);
  if OutFileName='' then OutFileName:=GetFName(InFileName);
  OutFileName := SetDefaultExt(OutFileName, '.WAV');

  writeln('Input file : ', InFileName);
  writeln('Output file: ', OutFileName);

  { Prepare XAudio decoder ------------------------------------------------ }

  { Create new decoder }
  Status := decoder_new(@decoder);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot init decoder (',
      xaudio_error_string(status),')');
    Halt(1);
  end;

  { Register the input module }
  Status:=file_input_module_register(@module);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot register file input module (',
      xaudio_error_string(status),')');
    Halt(2);
  end;

  Status:=decoder_input_module_register(decoder, @module);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot register decoder input module (',
      xaudio_error_string(status),')');
    Halt(3);
  end;


  { Set the notification functions (not really needed) }
  decoder^.notification_client.client           := decoder;
  decoder^.notification_client.notify_progress  := Notify_Progress;
  decoder^.notification_client.notify_debug     := Notify_Debug;
  decoder^.notification_client.notify_error     := Notify_Error;

  { output parameters }
  decoder^.output^.channels := XA_OUTPUT_STEREO;

  { open input }
  Move((InFileName+#0)[1], CharBuf, Length(InFileName)+1); { copy to PChar buf }

  Status:=decoder_input_new(decoder, @CharBuf, XA_DECODER_INPUT_AUTOSELECT);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot create input with decoder_input_new (',
      xaudio_error_string(status),')');
    Halt(4);
  end;

  Status:=decoder_input_open(decoder);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot open input with decoder_input_open (',
      xaudio_error_string(status),')');
    Halt(5);
  end;

  writeln;
  DecLen := decoder^.input^.duration;
  If decoder^.status^.info.mode=3 then Channels:=1
                                  else Channels:=2;
  write('MP3 file information: ',decoder^.status^.info.bitrate,'kbps, ');
  CASE decoder^.status^.info.mode OF
    0 : write('stereo, ');
    1 : write('joint-stereo, ');
    2 : write('dual-channel, ');
    3 : write('mono, ');
    else write('unknown mode, ');
  END;
  writeln(decoder^.status^.info.frequency,' Hz, ',
          DecLen div 60,'''', DecLen mod 60,'"');


  { Open output file ------------------------------------------------------ }

  Assign(OutF, OutFileName);
  ReWrite(OutF, 1);

  WavHead.FileHeader.ID     := 'RIFF';
  WavHead.FileHeader.Len    := 0; { unknown at this moment }
  WavHead.WavHeader         := 'WAVE';
  WavHead.Format.ID         := 'fmt ';
  WavHead.Format.Len        := SizeOf(tFormatChunk)-8;
  WavHead.Format.wFormatTag := 1;
  WavHead.Format.wChannels  := Channels;
  WavHead.Format.dwSamplesPerSec := decoder^.status^.info.frequency;
  WavHead.Format.dwAvgBytesPerSec := decoder^.status^.info.frequency*2*Channels;
  WavHead.Format.wBlockAlign:= 2*Channels;
  WavHead.Format.wBitsPerSample:=16;
  WavHead.Data.ID:='data';
  WavHead.Data.Len:=0; { unknown at this moment }

  BlockWrite(OutF, WavHead, SizeOf(WavHead));


  { Decode ---------------------------------------------------------------- }

  { start timer }
{$IFDEF ShowSpeed}
  Timer:=GetBiosTicks; // start timer
{$ENDIF}

  { Decode! }

  write('Decoding...    ');

  repeat
    IgnoreErr:=FALSE;

    { Decode (samples stored by XAudio internally) }
    Status := decoder_decode(decoder, NIL);

    { Calculate position }
    DecPos:=(decoder^.status.timecode.h*60*60) +
            (decoder^.status.timecode.m*60   ) +
            (decoder^.status.timecode.s      ) +
            (decoder^.status.timecode.f/100  );

    If DecLen>0 then write(#8#8#8#8#8#8,100/Real(DecLen)*DecPos:5:1,'%');


    { check status }
    If Status<>XA_SUCCESS then
    begin

      If Status=XA_ERROR_INVALID_FRAME then
      begin
        writeln(#13'Warning: Invalid frame encountered!');
        IgnoreErr:=TRUE; // this is a non-critical error
      end;

      If Status=XA_ERROR_TIMEOUT then
      begin
        { This should never happen with a file input, anyway... }
        writeln(#13'Warning: Timeout error while decoding!');
        IgnoreErr:=TRUE; // this is a non-critical error
      end;

    end else begin

      { store frame }
      BlockWrite(OutF, decoder^.output_buffer^.pcm_samples^,
                 decoder^.output_buffer.size);

    end;
  until (Status <> XA_SUCCESS) and (not IgnoreErr);

  writeln(#13'Decoding done                                   ');


  { Stop Timer }

{$IFDEF ShowSpeed}
  DecodingTime:=Round((GetBiosTicks-Timer)/TickFreq*1000);
  if DecodingTime=0
    then writeln('Zero milliseconds decoding time!?')
    else writeln('Decoding time: ',Real(DecodingTime)/1000:0:2, ' seconds, ',
                 1/(Real(DecodingTime)/1000)*DecPos:0:2,'x');
{$ENDIF}

  { Close output file ----------------------------------------------------- }

  WavHead.FileHeader.Len := FileSize(OutF)-8;
  WavHead.Data.Len := FileSize(OutF)-SizeOf(WavHead);
  Seek(OutF, 0);
  BlockWrite(OutF, WavHead, SizeOf(WavHead));

  Close(OutF);


  { Free XAudio decoder --------------------------------------------------- }

  Status:=decoder_delete(decoder);
  if Status <> XA_SUCCESS then
  begin
    writeln('*** ERROR: Cannot delete decoder (',
          xaudio_error_string(status),')');
    Halt(99);
  end;

end.
